/* ndk - [ pager.c ]
 *
 * Contains code to initialize the pager, and
 * to convert to and from linear and physical
 * address spaces.
 *
 * Please see:
 *   /src/memory.c
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */
#include "pager.h"

#include <stack.h>
#include <multiboot.h>

// see ndk/docs/memoryStacks.kwd for more details
typedef struct
{
   long total;
   Stack avail;
   Stack used;
} PageStack;

PageDirectory *sysPageDir;
PageStack pageStack;
uint32 totalMemory, totalPages;

// private functions...
void pagerCreatePageStack(void);
Boolean pagerTableExists(unsigned long tableNum);
void pagerCreateTable(unsigned long tableNum);
long pagerMapPage(unsigned long physAddr);   // TODO: remove, merge into new API
void *pagerLinToPhys(void *linearAddr);   // TODO: remove, merge into new API

ErrorCode pagerInit(void)
{
   PageTable *pTable;
   Pointer kbase, mbase;
   uint32 ksize, msize;
   long i, addr;
   uint32 pagesToAllocate;
   PhysicalAddress dummyPage;

   // find 4kb aligned address for page directory
   sysPageDir = (PageDirectory *) 0x00090000;

   // find 4kb aligned address for 1 page table
   pTable = (PageTable *) 0x00091000;

   // assign 1st page directory entry to page table
   sysPageDir->pageTable[0] = (long)pTable | PAGE_PRESENT | PAGE_READ_WRITE | PAGE_P3_ACCESS;
   // assign the last page table to the page directory!
   // This will aid in memory management (doc/memoryManagement.kwd)
   sysPageDir->pageTable[1023] = (long)sysPageDir | PAGE_PRESENT | PAGE_READ_WRITE | PAGE_P3_ACCESS;

   // clear the rest of the page directory
   for(i = 1; i < 1023; i++)
      sysPageDir->pageTable[i] = 0;

   // assign the page table simple linear address for the
   // first 4MBs of memory (later this should be changed to
   // map only used memory)
   //addr = 0;
   for(i = 0; i < 1024; i++)
   {
      pTable->pageAddr[i] = (i * 4096) | PAGE_PRESENT | PAGE_READ_WRITE | PAGE_P3_ACCESS;
      //printf("pTable[%d] = 0x%8x\n", i, pTable->pageAddr[i]);
   }

   // load CR3 with the address of our page directory
   printf("setting CR3: (0x%8x)\n", sysPageDir);
   setCR3(((getCR3() & 0xfff) | (long)sysPageDir));
   printf("CR3: 0x%x\n", getCR3());

   // and enable the paging bit in CR0
   printf("setting CR0: (0x%8x)\n", getCR0() | CR0_PAGING_ENABLE);
   setCR0((getCR0() | CR0_PAGING_ENABLE));
   printf("CR0: 0x%x\n", getCR0());

   pagerCreatePageStack();

   memoryGetKernelBase(&kbase);
   memoryGetKernelSize(&ksize);
   memoryGetModulesBase(&mbase);
   memoryGetModulesSize(&msize);
   printf("kernel base %x size %d\n", (long)kbase, (long)ksize);
   printf("moduels base %x size %d\n", (long)mbase, (long)msize);

   // allocate the first MB (all of conventional memory and also
   // allocate the space used by the kernel and modules
   //malloc(2*1024*1024);  // works...

   //malloc(1024*1024 + ksize + msize + 4096);
   // 40 == size of page stack!!!
//  pagesToAllocate = (1024*1024 + ksize + msize) / 4096 + 35;
   //pagesToAllocate = (2*1024*1024) / 4096 - 20;
//  for(i = 0; i < pagesToAllocate; i++)
//    pagerGetNewPhysical(&dummyPage);
   // This should work, but it doesn't... I suspect the rZero modules bss are using
   // memory past the kernel!!!
   // malloc(1024*1024 + ksize + msize);

   return ErrorNoError;
}

ErrorCode pagerGetPageSize(uint32 *page)
{
   *page = 4096;
   return ErrorNoError;
}

ErrorCode pagerGetNewPhysical(PhysicalAddress * addr)
{
   if(pageStack.avail.pointer == 0)
   {
      printf("Not enough physical memory!\n");
      return ErrorOutOfMemory;
   }

   // place the physical address onto the used stack, and
   // return it
   *addr = (PhysicalAddress) stackPop(&pageStack.avail);
   stackPush(&pageStack.used, (long)*addr);

   return ErrorNoError;
}

// allocates a page anywhere in linear space... useful?  I dunno...
// I have doubts that this really works, and whether it should even be
// implemented... it's basically a really crude malloc...
ErrorCode pagerGetNewLinear(LinearAddress * addr)
{
   PhysicalAddress pAddr;

   printf("pagerGetNewLinear()\n");

   // TODO: cleanup!  pagerMapPage is an old API as well..
   if(pagerGetNewPhysical(&pAddr) == ErrorNoError)
   {
      printf("Got physical %x\n", pAddr);
      *addr = (LinearAddress) pagerMapPage((uint32)pAddr);
      printf("Got linear %x\n", *addr);
      return ErrorNoError;
   }
   else
      return ErrorNotAvailable;
}

// TODO: error check if page exists!
ErrorCode pagerMapPhysicalToLinear(PhysicalAddress physicalAddr, LinearAddress linearAddr)
{
   long *pageMap = (long *)0xffc00000;
   long allocPageOffset = ((long)linearAddr / 4096);
   long pointer;

   printf("pagerMapPhysicalToLinear %x -> %x\n", (long)physicalAddr, (long)linearAddr);
   pointer = allocPageOffset;

   // make sure to create a page table if it doesn't exist!!!
   if(pointer % 1024 == 0 && pagerTableExists(pointer / 1024) == False)
   {
      printf("pagerMapPhysicalToLinear -> create new table\n");
      pagerCreateTable(pointer / 1024);
   }

   // put physical page into linear space
   //console().WriteFormat("tableNum %d, index %d, value 0x%8x\n", pointer / 1024, pointer, physicalAddr);
   pageMap[pointer] = (long)physicalAddr | PAGE_PRESENT | PAGE_READ_WRITE | PAGE_P3_ACCESS;
   //physAddr = (PhysicalAddress)(pointer*4096);
   return ErrorNoError;
}

ErrorCode pagerGetPhysical(LinearAddress lin, PhysicalAddress * addr)
{
   // TODO: clean this up... pagerLinToPhys is an old API!!!
   *addr = pagerLinToPhys((void *)lin);
}

ErrorCode pagerGetLinear(PhysicalAddress phys, LinearAddress * addr)
{
   // TODO: implement?
   return ErrorNotImplemented;
}

// PRIVATE...

void pagerCreatePageStack(void)
{
   module_t *mod;
   long addr;
   int i;
   long lastPageOfStack;

   printf("Usable memory: [Lower: %dkb, Upper %dkb]\n", multibootInfo->mem_lower,
          multibootInfo->mem_upper);

   // don't use mem_lower to calculate total memory... it'll return
   // 640k (ie, it doesn't include the bios).  Use 1024k (ie, 1MB)
   totalMemory = (multibootInfo->mem_upper + 1024) * 1024;
   totalPages = totalMemory / PAGE_SIZE;

   // kernel is currently always fixed @ 1MB mark, but could be any length
   // Use bssEnd to calculate total space used by kernel, remember to round
   // kernelEnd _UP_ to the nearest page.
   // TODO: UNUSED: move to memory.c
   //kernelStart = (1024 * 1024) / PAGE_SIZE;
   //kernelEnd = ((long)&bssEnd + PAGE_SIZE - ((long)&bssEnd & (PAGE_SIZE-1))) / PAGE_SIZE;

   // get the end address of the last module
   mod = (module_t *) multibootInfo->mods_addr;
   addr = (long)mod[multibootInfo->mods_count - 1].mod_end;

   // make sure it's at least dword aligned (rounding up!)
   addr += 4 - (addr & 3);
   printf("addr: 0x%8x\n", addr);

   /* The following calculations gave me much grief when
    * initially developing this code.  I had replaced addr
    * with dmaStack.base, or appStack.base when calculating
    * the used base pointer... although the two are exactly
    * equivalent (printf confirms), the calculation came
    * out wrong!  Just a word of warning...
    */
   printf("Creating page stack at: 0x%8x\n", addr);
   pageStack.total = totalPages;
   pageStack.avail.flags = STACK_EXPAND_DOWN;
   pageStack.avail.pointer = 0; //dmaStack.total - 1;
   pageStack.avail.base = addr;
   pageStack.used.flags = STACK_EXPAND_UP;
   pageStack.used.base = (addr + 4 * totalPages);
   pageStack.used.pointer = 0;

   /* Add all memory to the page stack except the first MB (conventional
    * memory -> to be maintained by ndk itself) and the memory that's
    * already allocated by the kernel and the page stack
    */
   lastPageOfStack = (pageStack.used.base / 4096) + 2;   // TODO: NOTE: off by one?
   for(i = pageStack.total; i > lastPageOfStack; i--)
   {
      //if(!pagerPageUsedByKernel(i))
      stackPush(&pageStack.avail, i * PAGE_SIZE);
   }
}

Boolean pagerTableExists(unsigned long tableNum)
{
   long *pgDir = (long *)0xfffff000;

   return pgDir[(unsigned long)tableNum] == 0 ? False : True;
}

void pagerCreateTable(unsigned long tableNum)
{
   long *pageDirectory = (long *)0xfffff000;
   long *pages = (long *)0xffc00000;
   long *pageTable;
   long i;

   printf("pager: creating new table #%d\n", tableNum);
   // get a page from the kernel
   //pageTable = (long *)memoryAllocUnmappedPage();
   pagerGetNewPhysical((PhysicalAddress *) & pageTable);
   printf("received page 0x%8x\n", pageTable);

   printf("Adding to directory\n");
   // and assign it to our page directory (must do before we write to it)
   pageDirectory[tableNum] = (long)pageTable | PAGE_PRESENT | PAGE_READ_WRITE | PAGE_P3_ACCESS;;

   // zero it out
   printf("And clearing\n");
   for(i = 0; i < 1024; i++)
      pages[tableNum * 1024 + i] = 0;  //pageTable[i] = 0;
}

long pagerMapPage(unsigned long physAddr)
{
   long *pages = (long *)0xffc00000;
   long allocPageOffset = (4 * 1024 * 1024) / 4096;
   long pointer;

   // simple way to find available slot
   pointer = allocPageOffset;
   while(pages[pointer++] != 0);

   // put physical page into linear space
   pages[pointer] = physAddr | PAGE_PRESENT | PAGE_READ_WRITE | PAGE_P3_ACCESS;
   return pointer * 4096;       //1024;
}

void *pagerLinToPhys(void *linearAddr)
{
   long *pages = (long *)0xffc00000;
   long linAddr = (long)linearAddr;
   long physAddr;
   long bottom4k;
   unsigned long pageTableNum;
   unsigned long pageIndex;

   // save the bottom 4k, and remove it from the linear address
   bottom4k = linAddr & 4095;
   linAddr -= bottom4k;

   // debugging info...
   printf("page aligned linAddr: 0x%8x\n", linAddr);

   // calculate which page table to look into
   pageTableNum = (unsigned long)linAddr / (4096 * 1024);
   //printf("table %d, index %d\n", pageTableNum, pageTableIndex);
   // Check if this page table exists!
   if(!pagerTableExists(pageTableNum))
   {
      printf("Page table doesn't exist for linear address 0x%8x\n", linAddr);
      return 0;
   }

   // calculate the offset into the page tables, and get our physical address
   pageIndex = (unsigned long)linAddr >> 12;
   physAddr = pages[(unsigned long)pageIndex] & 0xfffff000;

   // add the bottom 4k
   physAddr += bottom4k;

   return (void *)physAddr;
}
