/* ndk - [ messageQueue.c ]
 *
 * Basic message queue class for IPC
 *
 * (c)2004 dcipher / neuraldk
 *           www.neuraldk.org
 */

#include "messageQueue.h"
#include "expandingQueue.h"
#include "semaphore.h"
#include "mutex.h"
#include "queue.h"
#include <console.h>
#include <stdarg.h>

struct _MessageQueue
{
   union
   {
      ExpandingQueue expanding;
      Queue limited;
      void *queue;
   } queue;

   uint32 size;

   // general purpose get/put routines that *should* allow an
   // object oriented approach to the queues (ie, allow us to
   // transparently use either a return or expanding queue)
          ErrorCode(*get) (void *queue, Pointer *message);
          ErrorCode(*put) (void *queue, Pointer message);
          ErrorCode(*peek) (void *queue, Pointer *message);
   Mutex mutex;

   Semaphore sem;
};

struct _Message
{
   uint32 code;
   uint32 numArgs;
   MessageDestroyFunction destroy;
   MessageQueue responseQueue;

   Pointer arg[1];
};

ErrorCode messageQueueCreate(MessageQueue *msgq, uint32 size)
{
   MessageQueue mq = (MessageQueue)malloc(sizeof(struct _MessageQueue));

   if(size)
   {
      // create queue
      mq->get = (ErrorCode(*)(void *, Pointer *))queueGet;
      mq->put = (ErrorCode(*)(void *, Pointer))queueAdd;
      mq->peek = (ErrorCode(*)(void *, Pointer *))queuePeek;
      queueCreate(&mq->queue.limited, size);
   }
   else
   {
      // create expanding queue (ie, has no size)
      mq->get = (ErrorCode(*)(void *, Pointer *))expandingQueueGet;
      mq->put = (ErrorCode(*)(void *, Pointer))expandingQueueAdd;
      mq->peek = (ErrorCode(*)(void *, Pointer *))expandingQueuePeek;
      // todo: expanding queue to accept initial size?  Or no...?
      expandingQueueCreate(&mq->queue.expanding, 100);
   }
   mq->size = size;

   printf("Using queue of %x\n", mq->queue);

   semaphoreCreate(&(mq->sem), SemaphoreTypeFIFO, 0);
   mutexCreate(&(mq->mutex), MutexTypeFIFO);

   *msgq = mq;

   return ErrorNoError;
}

ErrorCode messageQueueSend(MessageQueue mq, Pointer message)
{
   ErrorCode ec = ErrorNoError;

   //mutexLock(mq->mutex, TimeoutInfinite);

   ec = mq->put((void *)mq->queue.queue, message);

   if(ec == ErrorNoError)
   {
      ec = semaphoreUp(mq->sem);
   }
   else
   {
      printf("Error sending message!\n");
   }

   //mutexUnlock(mq->mutex);

   return ec;
}

ErrorCode messageQueuePeek(MessageQueue mq, Pointer *message)
{
   ErrorCode ec;

   //mutexLock(mq->mutex, TimeoutInfinite);

   ec = mq->peek((void *)mq->queue.queue, message);

   //mutexUnlock(mq->mutex);

   return ec;
}

ErrorCode messageQueueReceive(MessageQueue mq, Timeout to, Pointer *message)
{
   ErrorCode ec;

   //mutexLock(mq->mutex, TimeoutInfinite);

   ec = semaphoreDown(mq->sem, to);
   if(ec == ErrorNoError)
   {
      ec = mq->get((void *)mq->queue.queue, message);
   }
   else
   {
      printf("error receiving message\n");
   }

   //mutexUnlock(mq->mutex);

   return ec;
}

ErrorCode messageQueueDestroy(MessageQueue *mq)
{
   if((*mq)->size)
   {
      queueDestroy(&((*mq)->queue.limited));
   }
   else
   {
      expandingQueueDestroy(&((*mq)->queue.expanding));
   }
   semaphoreDestroy(&((*mq)->sem));

   free(*mq);
   *mq = 0;

   return ErrorNoError;
}

ErrorCode messageCreate(Message * msg, MessageDestroyFunction destroy, MessageQueue responseQueue,
                        uint32 code, uint32 numArgs, ...)
{
   ErrorCode err = ErrorNoError;
   uint32 i;
   Message m;
   va_list ap;

   va_start(ap, numArgs);

   if(msg)
   {
      m = (Message) malloc(sizeof(struct _Message) + (numArgs - 1) * sizeof(Pointer));
      if(m)
      {
         m->code = code;
         m->numArgs = numArgs;
         m->destroy = destroy;
         m->responseQueue = responseQueue;

         for(i = 0; i < numArgs; i++)
         {
            m->arg[i] = va_arg(ap, Pointer);
         }
      }
      else
      {
         err = ErrorOutOfMemory;
      }
   }
   else
   {
      err = ErrorBadObject;
   }

   va_end(ap);

   *msg = m;

   return err;
}

ErrorCode messageGetArgument(Message msg, uint32 num, Pointer *arg)
{
   ErrorCode err = ErrorNoError;

   if(msg)
   {
      if(num <= msg->numArgs)
      {
         *arg = msg->arg[num - 1];
      }
      else
      {
         err = ErrorBadParam;
      }
   }
   else
   {
      err = ErrorBadObject;
   }

   return err;
}

ErrorCode messageGetArgumentCount(Message msg, uint32 *total)
{
   ErrorCode err = ErrorNoError;

   if(msg)
   {
      *total = msg->numArgs;
   }
   else
   {
      err = ErrorBadObject;
   }

   return err;
}

ErrorCode messageGetArgumentList(Message msg, Pointer *top, uint32 *total)
{
   ErrorCode err = ErrorNoError;

   if(msg)
   {
      *top = &msg->arg[0];
      *total = msg->numArgs;
   }
   else
   {
      err = ErrorBadObject;
   }

   return err;
}

ErrorCode messageGetCode(Message msg, uint32 *code)
{
   ErrorCode err = ErrorNoError;

   if(msg)
   {
      *code = msg->code;
   }
   else
   {
      err = ErrorBadObject;
   }

   return err;
}

ErrorCode messageRespond(Message source, Message response)
{
   ErrorCode err = ErrorNoError;

   if(source && source->responseQueue)
   {
      err = messageQueueSend(source->responseQueue, (Pointer)response);
   }
   else
   {
      err = ErrorBadObject;
   }

   return err;
}

ErrorCode messageDestroy(Message * msg)
{
   ErrorCode err = ErrorNoError;

   if(msg)
   {
      if((*msg)->destroy)
      {
         (*msg)->destroy((*msg));
      }
      free((*msg));
   }
   else
   {
      err = ErrorBadObject;
   }

   return err;
}
